#!/bin/sh
OUTPUT=stats
WORKDIR=`pwd`

if [ -f $OUTPUT ]; then
	echo $OUTPUT already exists!
	exit 1
fi

for d in $*; do
	if [ -d $d ]; then
		echo Processing $d

		NODOT=`echo $d | tr -d "."`
		CORES=`echo $d | cut -f2 -d"."`

		EMPTYOUT=`find $d -regex "$d/$NODOT\.[0-9].*out" | tail -1` 
		if [ -f "$EMPTYOUT" ]; then
			EMPTYSEC=`grep "Time in seconds" $EMPTYOUT | cut -f2 -d"=" | tr -d " "`
			EMPTYMOPS=`grep "Mop/s total" $EMPTYOUT | cut -f2 -d"=" | tr -d " "`
			echo $d,$EMPTYSEC,$EMPTYMOPS >> $OUTPUT
		fi

		RACEROUT=`find $d -regex "$d/$NODOT\.racer.*out" | tail -1` 
		if [ -f "$RACEROUT" ]; then
			RACERSEC=`grep "Time in seconds" $RACEROUT | cut -f2 -d"=" | tr -d " "`
			RACERMOPS=`grep "Mop/s total" $RACEROUT | cut -f2 -d"=" | tr -d " "`
			cd $d
			if [ ! -f "upctstats.out" ]; then
				upctstats > upctstats.out
			fi
			UPCTSTATS=`cat upctstats.out | grep "Total" | cut -f2 -d"[" | tr -d "\]"`
			NUMBARRIERS=`tail -1 upct.stat.0 | cut -f1 -d":"`
			cd $WORKDIR
			if [ ! -f $d/"commstats.out" ]; then
				grep COMM $RACEROUT | commstats > $d/commstats.out
			fi
			COMMSTATS=`cat $d/commstats.out | grep "Total" | cut -f2 -d"[" | tr -d "\]"`
			NUMBARRIERS=`expr $NUMBARRIERS + 1`
			echo $d.racer,$CORES,$NUMBARRIERS,$EMPTYSEC,$RACERSEC,$RACERMOPS,$UPCTSTATS,$COMMSTATS >> $OUTPUT
		else
			echo Cannot find $d/$NODOT.out
		fi

		TESTEROUT=`find $d -regex "$d/$NODOT\.tester.*out" | tail -1` 
		TESTERERR=`find $d -regex "$d/$NODOT\.tester.*err" | tail -1` 
		if [ -f "$TESTEROUT" ]; then
			TESTERSEC=`grep "Time in seconds" $TESTEROUT | cut -f2 -d"=" | tr -d " "`
			TESTERMOPS=`grep "Mop/s total" $TESTEROUT | cut -f2 -d"=" | tr -d " "`
			TESTERSUCC=`grep "REAL RACE" $TESTERERR | wc -l`
			echo $d.tester,$TESTERSEC,$TESTERMOPS,$TESTERSUCC >> $OUTPUT
		fi

	fi
done


