/*
 * allocator.c
 *
 *  Created on: Apr 11, 2011
 *      Author: parkcs
 */

#include <stdio.h>
#include <string.h>
#include "allocator.h"
#ifndef NO_UPC
#       include <upc.h>
#else
#       define MYTHREAD 0
#       define upc_global_exit(x)      exit(x)
#endif



void out_of_memory_exit(const char* file, int line, size_t bytes) {
    fprintf(stderr, "[%d] %s:%d Out of memory -- cannot allocate %lu bytes!\n",
            MYTHREAD, file, line, (unsigned long) bytes);
    upc_global_exit(77);
}

void new_mempool(memory_pool* mp) {
    if (mp) {
        memset(mp, 0, sizeof (memory_pool));
        /*mp->heap = NULL;
        mp->prev_pool = NULL;
        mp->top = 0;
        mp->limit = 0;*/
    }
}

void free_mempool(memory_pool* mp) {
    if (mp) {
        free(mp->heap);
        mp->heap = NULL;
        free_mempool(mp->prev_pool);
        mp->prev_pool = NULL;
        mp->top = 0;
        mp->limit = 0;
        free(mp);
    }
}

void recycle_mempool(memory_pool* mp) {
    if (mp) {
        free_mempool(mp->prev_pool);
        mp->prev_pool = NULL;
        mp->top = 0;
    }
}

#define max(a,b) ((a) >= (b) ? (a) : (b))

char* allocate(memory_pool* mp, size_t n) {
    if(n <= 0)
        return NULL;
    
    if (n > (mp->limit - mp->top)) {
        if (mp->limit) {
            memory_pool* prevpool = (memory_pool *) malloc(sizeof (memory_pool));
            if (!prevpool)
                out_of_memory(sizeof (memory_pool));
            memcpy(prevpool, mp, sizeof (memory_pool));
            mp->prev_pool = prevpool;
        }
        mp->top = 0;
        mp->limit = mp->limit < INIT_BLOCK_SIZE ? INIT_BLOCK_SIZE : mp->limit * 2;
        mp->limit = max(n, mp->limit);
        if (mp->limit > INIT_BLOCK_SIZE)
            fprintf(stderr, "[%d] Increasing memory by %lu bytes...\n", MYTHREAD, mp->limit);
        if (mp->limit > MAX_BLOCK_SIZE || !(mp->heap = malloc(mp->limit)))
            out_of_memory(mp->limit);
    }

    void* ret = mp->heap + mp->top;
    mp->top += (n + ALIGNMENT - 1) & ~(ALIGNMENT-1);
    //printf("mp: %p, top = %d\n", mp, mp->top);
    return ret;

}
#undef max
