/*
 * addr_pivl_skiplist.c
 *
 *  Created on: Sep 16, 2010
 *      Author: parkcs
 *  Rewritten: May 13, 2011
 *  Major changes: 
 *   - uses generalized skiplist with edge data embedded in skiplist nodes with
 *     statically-typed data
 *   - assumes left closed, right open intervals only (this is not a
 *     generalized interval skiplist implementation)
 */

#include <stdio.h>
#include <stdlib.h>
//#include <alloca.h>
#include "addr_interval_skiplist.h"

#if 0
static skiplist* endpoint_printer_edge_data_ref;
static pivl_pool* endpoint_printer_pivl_pool_ref;

void endpoint_printer(skiplist_node* x, int level) {
    printf("0x%"PRIx64" => ", x->key);
    pivl_set_print(endpoint_printer_pivl_pool_ref, (pivl_set*) x->data);
    for (int i = 0; i <= level; i++) {
        printf(" |%d: ", i);
        pivl_set_print(endpoint_printer_pivl_pool_ref, (pivl_set*) skiplist_find(
                endpoint_printer_edge_data_ref, (uint64_t) (&x->forward[i])));
    }
}
#endif

void addr_islist_print(addr_islist* sl) {
    pivl_pool_print(sl->interval_pool);
    addr_skiplist_print_pool = sl->interval_pool;
    addr_skiplist_print(sl->endpoints);
}

addr_islist* new_addr_islist(memory_pool* mp) {
    addr_islist* ret = (addr_islist*) allocate(mp, sizeof (addr_islist));
    ret->endpoints = addr_skiplist_new(mp);
    ret->interval_pool = pivl_pool_new(mp);

    return ret;
}

typedef struct nodeinfo_t {
    pivl I;
    addr_skiplist_node* start_node;
    int start_level;
    addr end_key;
    struct nodeinfo_t* next;
} nodeinfo;

#if 0
/* [private helper function for addr_islist_insert_endpoint]
 * removes markers along search path for I starting from start upto end
 * Note: Remove markers on endpoints as well if included in I
 * (the markers should be added later)
 */
void addr_islist_remove_markers(addr_islist* isl, nodeinfo* n) {
    addr_skiplist_node* x = n->start_node;

    for (int i = n->start_level; i >= 0; i--) {
        while (x->forward[i].link && x->forward[i].link->key < n->end_key) {
            // Remove I from edge
            pivl_set* is = x->forward[i].data;
            if (is)
                pivl_set_remove(isl->interval_pool, is, n->I);

            // Remove I from node if I contains x->key
            if (is_in_interval(x->key, n->I->range))
                pivl_set_remove(isl->interval_pool, x->data, n->I);

            x = x->forward[i].link;
        }
        // Here, x->key < n->end_key <= x->forward[i]->key
    }

    if (x->forward[0].link && x->forward[0].link->key == n->end_key) {
        pivl_set* is = &x->forward[0].data;
        if (is)
            pivl_set_remove(isl->interval_pool, is, n->I);
        if (is_in_interval(x->key, n->I->range))
            pivl_set_remove(isl->interval_pool, x->data, n->I);
        if (is_in_interval(x->forward[0].link->key, n->I->range))
            pivl_set_remove(isl->interval_pool, x->forward[0].link->data, n->I);
    }
}
#endif

#if 0
/* [private helper function for addr_islist_insert_endpoint]
 * adds markers along search path for I starting from start_node upto end_key
 * Note: "start_key" and "end_key" are all in "endpoints" (simplifies search)
 */
void addr_islist_add_markers(addr_islist* isl, nodeinfo* n) {
    addr_skiplist_node* x = n->start_node;

    // Add I to node x
    if (is_in_interval(x->key, n->I->range))
        pivl_set_add(isl->interval_pool, x->data, n->I);

    for (int i = n->start_level; i >= 0; i--) {
        while (x->forward[i].link && x->forward[i].link->key < n->end_key) {
            if (!x->forward[i].data) {
                x->forward[i].data = pivl_set_new(isl->interval_pool->mempool);
            }
            x->forward[i].data = pivl_set_add(isl->interval_pool,
                    x->forward[i].data, n->I);
            x->forward[i].link->data = pivl_set_add(isl->interval_pool,
                    x->forward[i].link->data, n->I);
            x = x->forward[i].link;
        }
        // Here, x->key < n->end_key <= x->forward[i]->key
        if (x->forward[i].link && x->forward[i].link->key == n->end_key) {
            if (!x->forward[i].data) {
                x->forward[i].data = pivl_set_new(isl->interval_pool->mempool);
            }
            x->forward[i]. data = pivl_set_add(isl->interval_pool,
                    x->forward[i].data, n->I);
            if (is_in_interval(x->forward[i].link->key, n->I->range))
                x->forward[i].link->data = pivl_set_add(isl->interval_pool,
                    x->forward[i].link->data, n->I);
            x = x->forward[i].link;
            break;
        }
    }
}

void set_max(nodeinfo* n) {
    if (n->I->range.start_addr <= n->start_node->key)
        return;

    addr_skiplist_node* x = n->start_node;
    uint64_t left = n->I->range.start_addr;

    for (int i = n->start_level; i >= 0; i--) {
        while (x->forward[i].link && x->forward[i].link->key <= left)
            x = x->forward[i].link;

        if (x->key == left) {
            n->start_node = x;
            n->start_level = i;
            return;
        }
    }
    abort();
}
#endif

/* [private helper function for addr_islist_insert]
 * Adds an endpoint of an interval to the interval skip list
 * adjusts existing markers if endpoint not already in list
 * Returns true if endpoint was not already in isl
 */
int addr_islist_insert_endpoint(memory_pool* tmp, addr_islist* isl, addr key) {
    addr_skiplist_node * update[MAX_SKIPLIST_LEVEL];
    addr_skiplist* sl = isl->endpoints;

    addr_skiplist_node* x = addr_skiplist_lookup(sl, key, update);
    if (x) // endpoint already in skiplist
        return FALSE;

    // Insert endpoint into skiplist 
    int level = addr_skiplist_new_level(sl, update);
    x = addr_skiplist_insert_node(sl, level, update);
    x->key = key;
    x->data = pivl_set_new(sl->mempool);

    // Place markers on edges leading out of x
    pivl_set* promoted = NULL;
    pivl_set* new_promoted = NULL;

    pivl_pool* ip = isl->interval_pool;
    pivl_iter it;
    
    int i;
    for (i = 0; i < level && x->FORWARD(i+1); ++i) {
        addr_interval upper_iv;
        upper_iv.start = key;
        upper_iv.end = x->FORWARD(i+1)->key;
        pivl_iter_init(&it, ip, update[i]->MARKERS(i));
        while (pivl_iter_hasnext(&it)) {
            pivl m = *pivl_iter_next(&it);
            int midx = pivl_iter_index(&it);
            if (contains_interval(m->range, upper_iv)) {
                new_promoted = pivl_set_addidx(tmp, new_promoted, midx);
                addr_skiplist_node* z = x;
                while (z != x->FORWARD(i+1)) {
                    pivl_set_removeidx(z->MARKERS(i), midx);
                    z = z->FORWARD(i);
                }
            } 
            else {
                x->MARKERS(i) = pivl_set_addidx(sl->mempool, x->MARKERS(i), midx);
            }
        }
        
        pivl_iter_init(&it, ip, promoted);
        while (pivl_iter_hasnext(&it)) {
            pivl m = *pivl_iter_next(&it);
            int midx = pivl_iter_index(&it);
            if (contains_interval(m->range, upper_iv)) {
                // Keep promoting
                addr_skiplist_node* z = x;
                while (z != x->FORWARD(i+1)) {
                    pivl_set_removeidx(z->MARKERS(i), midx);
                    z = z->FORWARD(i);
                }
            }
            else {
                x->MARKERS(i) = pivl_set_addidx(sl->mempool, x->MARKERS(i), midx);
                if (is_in_interval(x->FORWARD(i)->key, m->range)) {
                    x->FORWARD(i)->data = pivl_set_addidx(sl->mempool,
                            x->FORWARD(i)->data, midx);
                }
                pivl_iter_clear(&it);
            }
        }
        
        promoted = pivl_set_addall(tmp, promoted, new_promoted);
        new_promoted = NULL;
    }
    
    x->MARKERS(i) = pivl_set_clone(sl->mempool, update[i]->MARKERS(i));
    x->MARKERS(i) = pivl_set_addall(sl->mempool, x->MARKERS(i), promoted);
    
    pivl_iter_init(&it, ip, promoted);
    addr_skiplist_node* xfi = x->FORWARD(i);
    while (pivl_iter_hasnext(&it)) {
        pivl m = *pivl_iter_next(&it);
        if (is_in_interval(xfi->key, m->range))
            xfi->data = pivl_set_addidx(sl->mempool, xfi->data, 
                    pivl_iter_index(&it));
    }
    
    // Promote markers on edges leading into x
    promoted = NULL;
    new_promoted = NULL;

    for (i = 0; i < level; ++i) {
        addr_interval upper_iv;
        upper_iv.start = update[i+1]->key;
        upper_iv.end = key;
        
        pivl_iter_init(&it, ip, update[i]->MARKERS(i));
        while (pivl_iter_hasnext(&it)) {
            pivl m = *pivl_iter_next(&it);
            int midx = pivl_iter_index(&it);
            if (contains_interval(m->range, upper_iv)) {
                new_promoted = pivl_set_addidx(tmp, new_promoted, midx);
                addr_skiplist_node* z = update[i+1];
                while (z != x) {
                    pivl_set_removeidx(z->MARKERS(i), midx);
                    z = z->FORWARD(i);
                }
            }
        }
        
        pivl_iter_init(&it, ip, promoted);
        while (pivl_iter_hasnext(&it)) {
            pivl m = *pivl_iter_next(&it);
            int midx = pivl_iter_index(&it);
            if (contains_interval(m->range, upper_iv)) {
                // keep promoting
                addr_skiplist_node* z = update[i+1];
                while (z != x) {
                    pivl_set_removeidx(z->MARKERS(i), midx);
                    z = z->FORWARD(i);
                }
            }
            else {
                update[i]->MARKERS(i) = pivl_set_addidx(sl->mempool,
                        update[i]->MARKERS(i), midx);
//                if (is_in_interval(update[i]->key, m->range)) {
                    update[i]->data = pivl_set_addidx(sl->mempool,
                            update[i]->data, midx);
//                }
                pivl_iter_clear(&it);
            }
        }
        
        promoted = pivl_set_addall(tmp, promoted, new_promoted);
        new_promoted = NULL;
    }
    
    update[i]->MARKERS(i) = pivl_set_addall(sl->mempool,
                update[i]->MARKERS(i), promoted);
    update[i]->data = pivl_set_addall(sl->mempool, update[i]->data, promoted);
    
    /*pivl_iter_init(&it, ip, promoted);
    while (pivl_iter_hasnext(&it)) {
        pivl m = pivl_iter_next(&it);
        if (is_in_interval(update[i]->key, m->range)) {
            update[i]->data = pivl_set_addidx(sl->mempool, update[i]->data,
                    pivl_iter_index(&it));
        }
    }*/
    
    for (i = 0; i <= level; ++i)
        x->data = pivl_set_addall(sl->mempool, x->data, x->MARKERS(i));
  
    return TRUE;
#if 0
    // removes intervals from update[*]
    skiplist_node* L = update[level];
    skiplist_node* R = update[level]->forward[level];
    for (int i = 0; i <= level; i++) {
        /* We're searching through the markers on the forward edge
         * Thus, each x corresponds to an interval
         */
        pivl_set* markers = (pivl_set*) skiplist_find(isl->edge_data,
                (uint64_t)&(update[i]->forward[i]));
        if (markers) {
            pivl_iter it;
            pivl_iter_init(&it, isl->interval_pool, markers);
            while (pivl_iter_hasnext(&it)) {
                pivl x = *pivl_iter_next(&it);
                nodeinfo* old_affected = affected;
                affected = alloca(sizeof (nodeinfo));
                affected->I = x;
                affected->start_node = L;
                affected->start_level = i;
                affected->end_key = (R && (uint64_t) R->key <= x->range.end_addr) ?
                        (uint64_t) R->key : x->range.end_addr;
                affected->next = old_affected;
                set_max(affected);
                addr_islist_remove_markers(isl, affected);
            }
        }
    }

    // add back intervals from update[*]
    while (affected) {
        addr_islist_add_markers(isl, affected);
        affected = affected->next;
    }
#endif
}

/* Private helper function for addr_islist_insert
 * Inserts "key" on edges and nodes from "start" (at "level") to "end",
 * descending levels as necessary 
 */
void addr_islist_insert_until(memory_pool* mp, int level, int Iidx,
        addr_skiplist_node* start_node, addr end_key) {
    addr_skiplist_node* x = start_node;
    int i;
    for (i = level; i >= 0; --i) {
        addr_skiplist_node* xf;
        while ((xf = x->FORWARD(i)) && (xf->key <= end_key)) {
            x->MARKERS(i) = pivl_set_addidx(mp, x->MARKERS(i), Iidx);
            x->data = pivl_set_addidx(mp, x->data, Iidx);
            x = xf;
        }
        // Either xf == NULL or xf->key > end_key
    }
#if 0
    while(x && (level >= 0) && (x->key < end_key)) {
        addr_skiplist_node* xf;
        while((xf = x->FORWARD(level)) && (xf->key < end_key)) {
            x->MARKERS(level) = pivl_set_addidx(mp, x->MARKERS(level), Iidx);
            x->data = pivl_set_addidx(mp, x->data, Iidx);
            x = xf;
        }
        --level;
    }
#endif
}

/* Inserts an interval into the IS-list 
 * Param tmp: need to pass a temporary memory pool used for interal data
 *            required only during the lifetime of function
 *        sl: IS-list to insert into
 *       key: interval to insert 
 */
void addr_islist_insert(memory_pool* tmp, addr_islist* sl, pivl key) {
    addr A = key->range.start;
    addr B = key->range.end;
    int ret;
    int Iidx = pivl_insert(sl->interval_pool, key);
#if 0 && defined(THRILLE_DEBUG)
    printf("Inserting key ");
    pivl_print(key);
    printf(" with index %d\n", Iidx);
#endif
    // insert key->start_addr and key->end_addr if necessary
    ret = addr_islist_insert_endpoint(tmp, sl, A);
#if 0 && defined(THRILLE_DEBUG)
    if (ret)
        addr_islist_print(sl);
#endif
    ret = addr_islist_insert_endpoint(tmp, sl, B);
#if 0 && defined(THRILLE_DEBUG)
    if (ret)
        addr_islist_print(sl);
#endif

    // add markers for [key->start_addr, key->end_addr)
    // Make sure markers on nodes are added appropriately
    
    memory_pool* mp = sl->endpoints->mempool;
    addr_skiplist_node* x = sl->endpoints->head;
    addr end_key = B;
    
    int i;
    for (i = sl->endpoints->level; i >= 0; i--) {
        addr_skiplist_node* xf;
        while ((xf = x->FORWARD(i)) && xf->key < A)
            x = xf;

        if (xf && (xf->key < end_key)) {
#if 0 && defined(THRILLE_DEBUG)
            printf("Insert from 0x%llx until 0x%llx...\n", xf->key, end_key);
#endif
            addr_islist_insert_until(mp, i, Iidx, xf, end_key);
            end_key = xf->key;
#if 0 && defined(THRILLE_DEBUG)
            printf("end_key = 0x%llx\n", end_key);
            addr_islist_print(sl);
#endif
        }
    }
    
#if 0
    nodeinfo n;
    n.I = key;
    n.start_node = sl->endpoints->head;
    n.start_level = sl->endpoints->level;
    n.end_key = key->range.end_addr;
    set_max(&n);
    addr_islist_add_markers(sl, &n);
#endif
}

/* Returns all intervals that contain the point key
 * Param tmp: memory pool to allocate set of overlapping intervals
 */
pivl_set * addr_islist_overlap_point(memory_pool* tmp, addr_islist* sl, addr key) {
    pivl_set* ret = NULL;
    addr_skiplist_node* x = sl->endpoints->head;
    int i;
    for (i = sl->endpoints->level; i >= 0; i--) {
        addr_skiplist_node* xf;
        while ((xf = x->FORWARD(i)) && xf->key < key) {
            x = xf;
        }
        // At this point x->key < key <= x->FORWARD(i)->key
        if (xf && xf->key == key) {
            ret = pivl_set_addall(tmp, ret, xf->data);
        }
        else {
            // all intervals in x->MARKERS(i) contain [x->key, xf->key) which in
            // turn contains key
            ret = pivl_set_addall(tmp, ret, x->MARKERS(i));
        }
    }

    return ret;
}

/* Returns all intervals that overlap with interval I
 * Param mp: memory pool to allocate set of overlapping intervals
 * 
 * Note: Some optimizations can be done if all intervals in islist are
 * of the form [a,b) (as in the case with memory addresses)
 * -- only need to scan for the left endpoint and add all data following
 * level 0 forward edges until right endpoint
 */
pivl_set * addr_islist_overlap_interval(memory_pool* tmp,
        addr_islist* sl, addr_interval * I) {
    pivl_set* ret = NULL;
    addr_skiplist_node* x = sl->endpoints->head;

    addr start = I->start;
    addr_skiplist_node* xf = NULL;
    int i;
    for (i = sl->endpoints->level; i >=0; --i) {
        while ((xf = x->FORWARD(i)) && xf->key < start) {
            x = xf;
        }
        // At this point x->key < start <= x->FORWARD(i)->key
        if (xf && xf->key == start)
            break;

        // all intervals in x->MARKERS(i) contain [x->key, xf->key) which in
        // turn overlaps with I
        ret = pivl_set_addall(tmp, ret, x->MARKERS(i));
    }
    
    // At this point xf->key >= start
    // Now add all markers on nodes up to I->end
    addr end = I->end;
    while (xf && xf->key < end) {
        ret = pivl_set_addall(tmp, ret, xf->data);
        xf = xf->FORWARD(0);
    }
    
#if 0
    for (int i = sl->endpoints->level; i > 0; i--) {
        // First search for left endpoint a of I=[(a,b)]
        while (x->forward[i] && comp(x->forward[i]->key, I->start_addr) < 0) {
            x = x->forward[i];
        }
        if (x->forward[i] && comp(x->forward[i]->key, I->start_addr) == 0) {
            x = x->forward[i];
            if (is_left_closed(*I)) {
                ret = pivl_set_addall(tmp, ret, (pivl_set*) x->data);
            }
        }
        // At this point, x->key < a <= x->forward[i]->key, thus, all
        // intervals on forward edges of level i up to b is overlapping
        skiplist_node* y = x;
        while (y->forward[i] && comp(y->forward[i]->key, I->end_addr) < 0) {
            ret = pivl_set_addall(tmp, ret, (pivl_set*) skiplist_find(
                    sl->edge_data, (uint64_t)&(y->forward[i])));
            y = y->forward[i];
        }
        ret = pivl_set_addall(tmp, ret, (pivl_set*) skiplist_find(
                sl->edge_data, (uint64_t)&(y->forward[i])));
    }

    // Special case for level 0: inclusion of endpoints dependent on type
    while (x->forward[0] && comp(x->forward[0]->key, I->start_addr) < 0) {
        x = x->forward[0];
    }
    // x->key < a <= x->forward[0]->key
    if (x->forward[0] && comp(x->forward[0]->key, I->start_addr) == 0) {
        x = x->forward[0];
        if (is_left_closed(*I))
            ret = pivl_set_addall(tmp, ret, (pivl_set*) x->data);
    }

    // add all intervals on forward edges and nodes until we reach b
    while (x->forward[0] && comp(x->forward[0]->key, I->end_addr) < 0) {
        ret = pivl_set_addall(tmp, ret, (pivl_set*) skiplist_find(
                sl->edge_data, (uint64_t)&(x->forward[0])));
        ret = pivl_set_addall(tmp, ret, (pivl_set*) x->forward[0]->data);
        x = x->forward[0];
    }
    // x->key < b <= x->forward[0]->key
    ret = pivl_set_addall(tmp, ret, (pivl_set*) skiplist_find(
            sl->edge_data, (uint64_t)&(x->forward[0])));
    if (x->forward[0] && comp(x->forward[0]->key, I->end_addr) == 0 && is_right_closed(*I))
        ret = pivl_set_addall(tmp, ret, (pivl_set*) x->forward[0]->data);
#endif
    return ret;
}

