/*
 * addr_interval_skiplist.h
 *
 *  Created on: Sep 16, 2010
 *      Author: parkcs
 */

#ifndef ADDR_INTERVAL_SKIPLIST_H_
#define ADDR_INTERVAL_SKIPLIST_H_

#include <stdbool.h>
#include "addr_skiplist.h"
#include "interval_set.h"
#include "../allocator.h"

/* Interval Skip List (from Hanson92: The Interval Skip List: A Data Structure for
 * Finding All Intervals That Overlap a Point
 *
 * Nodes and edges have associated set of intervals (Map<void*, interval_set>)
 * Keep set of intervals in cache (Map<hash_code, interval_set>)
 *
 * Structure:
 * - Each node in the skip list represents an end point of intervals in is-list
 * - Each edge in the forward list represents an interval (n1, n2)
 * - Nodes and edges are annotated with intervals in is-list that contain them
 *
 * Invariant:
 * - Only the maximal interval in forward list is annotated
 *   i.e. if (a,b) \subset (c,d) \subset I, only (c,d) is marked with I
 *
 * */


typedef struct addr_islist_t {
    addr_skiplist* endpoints;           // with markers on nodes and edges
    pivl_pool* interval_pool;
} addr_islist;


pivl_set* addr_islist_overlap_point(memory_pool*, addr_islist* sl, addr key);
pivl_set* addr_islist_overlap_interval(memory_pool*, addr_islist* sl, addr_interval* key);
void addr_islist_insert(memory_pool*, addr_islist* sl, pivl key);
addr_islist* new_addr_islist(memory_pool*);
void addr_islist_print(addr_islist* sl);


#endif /* ADDR_INTERVAL_SKIPLIST_H_ */
