/*
 ============================================================================
 Name        : interval_tree.c
 Author      : 
 Version     :
 Copyright   : Your copyright notice
 Description : Hello World in C, Ansi-style
 ============================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>

#include "addr_skiplist.h"
#include "addr_interval_skiplist.h"

#define NUM_INSERTS	500000
#define NUM_SEARCHES	1000000

long long
timeval_diff(struct timeval *difference,
             struct timeval *end_time,
             struct timeval *start_time
            )
{
  struct timeval temp_diff;

  if(difference==NULL)
  {
    difference=&temp_diff;
  }

  difference->tv_sec =end_time->tv_sec -start_time->tv_sec ;
  difference->tv_usec=end_time->tv_usec-start_time->tv_usec;

  /* Using while instead of if below makes the code slightly more robust. */

  while(difference->tv_usec<0)
  {
    difference->tv_usec+=1000000;
    difference->tv_sec -=1;
  }

  return 1000000LL*difference->tv_sec+
                   difference->tv_usec;

}

#include "lockset.h"

int main(void) {

	printf("size of lock: %lu\n", sizeof(lock));
	printf("size of lockpool_item_t: %lu\n", sizeof(lockpool_item_t));
	printf("size of lockpool: %lu\n", sizeof(lockpool));
	printf("size of lockset: %lu\n", sizeof(lockset));

	ls_newpool();

	lockpool_item_t A, B;
	A[0] = 123;
	A[1] = 234;

	printf("A[0] = %llu\n", A[0]);
	printf("A[1] = %llu\n", A[1]);
	printf("B[0] = %llu\n", B[0]);

	ls_freepool();

	return 0;

	struct timeval t1, t2, tdiff;

	addr_interval intervals[] = {{ 0x10, 0x20, OPEN },
	 { 0x20, 0x30, RIGHT_OPEN},
	 { 0x16, 0x50, CLOSED },
	 { 0x60, 0x80, LEFT_OPEN },
	 { 0x20, 0x70, RIGHT_OPEN },
	 { 0x15, 0x40, CLOSED },
	 { 0x55, 0x55, CLOSED },
	 { 0x100, 0x200, OPEN },
	 { 0x90, 0x100, CLOSED },
	 { 0x300, 0x400, OPEN },
	 { 0x500, 0x600, OPEN },
	 { 0x800, 0x1000, OPEN },
	 { 0x0, 0x0, OPEN}};

	void* search_list[] = {0x10, 0x18, 0x20, 0x28,
			0x30, 0x16, 0x33, 0x50,
			0x60, 0x70, 0x80, 0x48,
			0x15, 0x2a, 0x40, 0x55, 0x150, 0x9a, 0x350, 0x550, 0x900, 0x52, 0x0};
	int search_result[] = {0, 3, 4, 4,
			3, 3, 3, 2,
			1, 1, 1, 2,
			2, 4, 3, 2, 1, 1, 1, 1, 1, 1, 0};

	addr_interval search_intervals[] = {{ 0x10, 0x20, OPEN },
	 { 0x20, 0x30, RIGHT_OPEN},
	 { 0x16, 0x50, CLOSED },
	 { 0x60, 0x80, LEFT_OPEN },
	 { 0x20, 0x70, RIGHT_OPEN },
	 { 0x15, 0x40, CLOSED },
	 { 0x5, 0x10, CLOSED },
	 { 0x80, 0x100, LEFT_OPEN },
	 { 0x50, 0x60, LEFT_OPEN },
	 { 0x25, 0x45, OPEN },
	 { 0x41, 0x41, CLOSED },
	 { 0x41, 0x42, CLOSED },
	 { 0x0, 0x0, OPEN}};

	int search_result2[] = {3, 4, 5, 2, 6, 5, 0, 1, 2, 4, 2, 2, 0};

	addr_islist* isl = new_addr_islist();

	int i = -1;
	while(intervals[++i].start_addr) {
		addr_islist_insert(isl, intervals+i, (void*)10);
		printf("After inserting %p = %c%p,%p%c\n", intervals+i,
				is_left_closed(intervals[i]) ? '[' : '(',
				intervals[i].start_addr, intervals[i].end_addr,
				is_right_closed(intervals[i]) ? ']' : ')');
		addr_islist_print(isl);
	}

	/*addr_islist_print(isl);
	printf("After inserting %p = [0x20, 0x30)\n", &i2);
	addr_islist_insert(isl, &i2, (void*)10);
	addr_islist_print(isl);
	printf("After inserting %p = [0x16, 0x50]\n", &i3);
	addr_islist_insert(isl, &i3, (void*)10);
	addr_islist_print(isl);*/

	i = 0;
	while (search_list[i]) {
		interval_set* result = addr_islist_overlap_point(isl, search_list[i]);
		if(interval_set_size(result) == search_result[i]) {
			printf("Result for search on point %p matches.\n", search_list[i]);
		}
		else {
			printf("Result for search on point %p: |", search_list[i]);
			interval_set_printer(result);
			printf("| != %d (expected).\n", search_result[i]);
		}
		delete_interval_set(result);
		i++;
	}

	i = 0;
	while (search_intervals[i].start_addr) {
		printf("Result for search on ");
		interval_printer(search_intervals+i);
		interval_set* result = addr_islist_overlap_interval(isl, search_intervals+i);
		if(interval_set_size(result) == search_result2[i]) {
			printf(" matches.\n");
		}
		else {
			printf(": |");
			interval_set_printer(result);
			printf("| != %d (expected).\n", search_result2[i]);
		}
		delete_interval_set(result);
		i++;
	}

	printf("0x30 in intervals: ");
	interval_set_printer(addr_islist_overlap_point(isl, (void*)0x30));
	printf("\n");

	interval_printer(search_intervals+4);
	printf(" overlaps with intervals: ");
	interval_set_printer(addr_islist_overlap_interval(isl, search_intervals+4));
	printf("\n");


//	printf("Endpoints:\n");
//	skiplist_print(isl->endpoints);
//	printf("Markers:\n");
//	skiplist_print(isl->edge_data);

/*	skiplist* sl = new_addr_skiplist();
	skiplist_print(sl);
	skiplist_insert(sl, 10, 1);
	skiplist_print(sl);
	skiplist_insert(sl, 19, 1);
	skiplist_print(sl);
	skiplist_insert(sl, 1, 1);
	skiplist_print(sl);
	skiplist_insert(sl, 27, 1);
	skiplist_print(sl);

	gettimeofday(&t1, NULL);
	for(int i = 0; i < NUM_INSERTS; i++)
		skiplist_insert(sl, (void*)((uintptr_t)rand()), (void*)2L);
	gettimeofday(&t2, NULL);
	printf("%d inserts in %lld usec.\n", NUM_INSERTS,
			timeval_diff(&tdiff, &t2, &t1));

	gettimeofday(&t1, NULL);
	for(int i = 0; i < NUM_SEARCHES; i++)
		skiplist_find(sl, (void*)((uintptr_t)rand()));
	gettimeofday(&t2, NULL);
	printf("%d searches in %lld usec.\n", NUM_SEARCHES,
			timeval_diff(&tdiff, &t2, &t1));

	//skiplist_print(sl);
*/

	return EXIT_SUCCESS;
}
