/*
 * race_trie.c
 *
 *  Created on: Nov 3, 2010
 *      Author: parkcs
 */

#include "race_trie.h"
#include "allocator.h"

race_trie_node* race_trie_new(memory_pool* mp)
{
	race_trie_node* ret = (race_trie_node*) allocate(mp, sizeof(race_trie_node));
	ret->children = NULL;
	ret->data = NULL;
	return ret;
}

void race_trie_free(race_trie_node* rt)
{
	/*
	// recursively free children;
	if(rt) {
		trie_ch_node* c = rt->children;
		while(c) {
			race_trie_free(c->node);
			trie_ch_node* tmp = c;
			c = c->next;
			free(tmp);
		}
		free(rt);
	}
	*/
}

trie_ch_node* trie_ch_new(memory_pool* mp, lock l, race_trie_node* n) {
	trie_ch_node* ret = (trie_ch_node*) allocate(mp, sizeof(trie_ch_node));
	ret->lock = l;
	ret->next = 0;
	ret->node = n;
	return ret;
}

race_trie_node* race_trie_create(memory_pool* mp, lock_sorted* ls, struct access_info_t* ai)
{
	race_trie_node* head = race_trie_new(mp);
	race_trie_node* current = head;

	for(int i = 0; i < ls->length; i++) {
		current->children = trie_ch_new(mp, ls->data[i], race_trie_new(mp));
		current = current->children->node;
	}

	current->data = (access_info*) allocate(mp, sizeof(access_info));
	memcpy(current->data, ai, sizeof(access_info));

	return head;
}

/*void race_trie_add(race_trie_node* node, lock_sorted* ls, int index, struct access_info_t* ai)
{
	if(index < ls->length) {
		trie_ch_node* p = 0;
		trie_ch_node* c = node->children;
		while (c) {
			if(c->lock < ls->data[index]) {
				p = c;
				c = c->next;
			}
			else if(c->lock == ls->data[index]) {
				race_trie_add(c->node, ls, index+1, ai);
				return;
			}
			else	// lock not in trie
				break;
		}
		c = trie_ch_new(ls->data[index], race_trie_new());
		if(p) {
			c->next = p->next;
			p->next = c;
		}
		else {
			c->next = node->children;
			node->children = c;
		}
		race_trie_add(c->node, ls, index+1, ai);
	}
	if(node->data) {
		if(is_weaker_access(*ai, *(node->data)))
			node->data = ai;
	}
	else {
		node->data = ai;
	}
}*/

/* Private helper function
 * DFS of trie to check for weaker accesses than acc
 * only traverses edges with labels in acc's lockset
 */
static access_info* _check_weaker(race_trie_node* node,
		access_info_full* acc, int lock_index)
{
	if(node->data) {
		if(is_weaker_access(*(node->data), acc->ai))
			return node->data;
	}

	trie_ch_node* c = node->children;
	while(c && lock_index < acc->ls.length) {
		access_info* weaker;
		switch(lock_comp(&c->lock, &acc->ls.data[lock_index])) {
		case -1:	// child has extra lock
			c = c->next;
			break;
		case 0:		// child has common lock
			if ((weaker = _check_weaker(c->node, acc, ++lock_index)))
				return weaker;
			c = c->next;
			break;
		case 1:		// acc has extra lock, skip over
			++lock_index;
			break;
		default:	// should not reach here
			return 0;
		}
	}

	return 0;
}

/* Returns a node in trie (and its children) that is weaker than acc */
access_info* race_trie_check_weaker(race_trie_node* node, access_info_full* acc)
{
	return _check_weaker(node, acc, 0);
}

/* Private helper function
 * DFS of trie to check for racing accesses than acc
 * only traverses edges with labels not in acc's lockset
 */
static access_info* _check_race(race_trie_node* node,
		access_info_full* acc, int lock_index)
{
	if(node->data) {
		if(is_racing_access(*(node->data), acc->ai))
			return node->data;
	}

	trie_ch_node* c = node->children;
	while(c) {
		int comp = lock_index < acc->ls.length ?
				lock_comp(&c->lock, &acc->ls.data[lock_index]) : -1;
		access_info* race;
		switch(comp) {
		case -1:	// child has extra lock
			if ((race = _check_race(c->node, acc, lock_index)))
				return race;
			break;
		case 0:		// child has common lock
			++lock_index;
			break;
		case 1:		// acc has extra lock
			if ((race = _check_race(c->node, acc, ++lock_index)))
				return race;
			break;
		default:	// should not reach here
			return 0;
		}
		c = c->next;
	}

	return 0;
}

/* Returns a node in trie (and its children) that races with acc */
access_info* race_trie_check_race(race_trie_node* node, access_info_full* acc)
{
	return _check_race(node, acc, 0);
}

static void _merge(memory_pool* mp,
		race_trie_node* node, access_info_full* acc, int index)
{
	if(index < acc->ls.length) {
		trie_ch_node* p = 0;
		trie_ch_node* c = node->children;
		while (c) {
			if(c->lock < acc->ls.data[index]) {
				p = c;
				c = c->next;
			}
			else if(c->lock == acc->ls.data[index]) {
				_merge(mp, c->node, acc, index+1);
				return;
			}
			else	// lock not in trie
				break;
		}
		c = trie_ch_new(mp, acc->ls.data[index], race_trie_new(mp));
		if(p) {
			c->next = p->next;
			p->next = c;
		}
		else {
			c->next = node->children;
			node->children = c;
		}
		_merge(mp, c->node, acc, index+1);
		return;
	}
	if(node->data) {
		if (acc->ai.access == WRITE) {	// newer write overwrites
			if (node->data->thread != acc->ai.thread)
				node->data->thread =  MULTIPLE_THREADS;
			node->data->access = acc->ai.access;
			node->data->sig = acc->ai.sig;
			node->data->phase = acc->ai.phase;
		}
		else if (node->data->access == WRITE) {
			if (node->data->thread != acc->ai.thread)
				node->data->thread =  MULTIPLE_THREADS;
		}
		else {	// both READs, newer read overwrites
			if (node->data->thread != acc->ai.thread)
				node->data->thread =  MULTIPLE_THREADS;
			node->data->sig = acc->ai.sig;
			node->data->phase = acc->ai.phase;
		}
	}
	else {
		node->data = (access_info*) allocate(mp, sizeof(access_info));
		memcpy(node->data, &acc->ai, sizeof(access_info));
	}

}

void race_trie_merge(memory_pool* mp, race_trie_node* node, access_info_full* acc)
{
	_merge(mp, node, acc, 0);
}

static void _remove_stronger(race_trie_node* node, access_info_full* acc, int index)
{
	if(index >= acc->ls.length) {	// All children including this node may be stronger
		if(node->data && is_weaker_access(acc->ai, *(node->data)))
			node->data = 0;
	}

	trie_ch_node* c = node->children;
	while(c) {
		int comp = index < acc->ls.length ?
				lock_comp(&c->lock, &acc->ls.data[index]) : -1;
		access_info* race;
		switch(comp) {
		case -1:	// child has extra lock
			_remove_stronger(c->node, acc, index);
			break;
		case 0:		// child has common lock
			_remove_stronger(c->node, acc, ++index);
			break;
		case 1:		// acc has extra lock
		default:
			return;
		}
		c = c->next;
	}
}

void race_trie_remove_stronger(race_trie_node* node, access_info_full* acc)
{
	_remove_stronger(node, acc, 0);
}
