/*
 * upct_misc.c -- implementation of arch dependent functions
 *
 */

#include "upct_types.h"
#include <stdio.h>
#include <string.h>

#define RACE_INFO_FILENAME	"upct.race"
#define MAX_FILENAME_SIZE	256
#define MAX_LINE_SIZE	1024

#define STAT_FILENAME		"upct.stat"

/* Copy local string to shared 
 * src must be null terminated */

/*
void ls_strcpy(const char* src, shared [] const char* dst) {
    int n = strlen(src) + 1;
    shared [] char* tmp = (shared [] char*) upc_alloc(n);
    upc_memput(tmp, src, n);
    dst = tmp;
}
 */

/* Copy shared string to local */
/*
void sl_strcpy(shared [] const char* src, const char* dst) {
    int n = strlen(src) + 1;
    shared [] char* tmp = (shared [] char*) upc_alloc(n);
    upc_memput(tmp, src, n);
    dst = tmp;
}
 */

#if 0
void print_lock_list(shared void * ls) {
	shared lock_addr * mylock = (shared lock_addr *) ls;
	fprintf(stderr, "%"PRIx64"",(*mylock));
}

void print_race_info(shared race_info_t * info) {
	fprintf(stderr, "Raceinfo(Addr:%"PRIx64", ", info->addr);
	fprintf(stderr, "Thr:%d, ", info->thread);
	fprintf(stderr, "Phase:%d, ", info->phase);
	fprintf(stderr, "Access:%s, ", info->access == READ ? "R" : "W");
	fprintf(stderr, "LocksetSize:%d, ", list_size(info->locklist));
	fprintf(stderr, "Lockset:");
	list_print(info->locklist, &print_lock_list);
	fprintf(stderr, ", ");
	fprintf(stderr, "Sig:%s, ", info->sig);
	fprintf(stderr, "RaceReported:%s, ", info->race_reported == REPORTED ? "Y": "N");
	fprintf(stderr, "Owned:%s", info->owned == TRUE ? "Y": "N");
	fprintf(stderr, ")");
}
#endif

void print_race_info_ns(race_info_t * info) {
	fprintf(stderr, "Raceinfo(Addr:0x%"PRIx64"-%"PRIx64", ", info->addr[0], info->addr[1]);
	fprintf(stderr, "Thr:%d, ", info->thread);
	fprintf(stderr, "Phase:%d, ", info->phase);
	fprintf(stderr, "Access:%s, ", info->access == READ ? "R" : "W");
	//fprintf(stderr, "LocksetSize:%d, ", list_size(info->locklist));
	//fprintf(stderr, "Lockset:");
	//list_print(info->locklist, &print_lock_list);
	//fprintf(stderr, ", ");
	fprintf(stderr, "Sig:%s, ", info->sig);
	fprintf(stderr, "RaceReported:%s", info->race_reported == REPORTED ? "Y": "N");
	fprintf(stderr, "Owned:%s", info->owned == TRUE ? "Y": "N");
	fprintf(stderr, ")");
}

char fname[MAX_FILENAME_SIZE];


void write_race_info(FILE* fp, addr_interval* i, access_info* a, int myt) {
	fprintf(fp, "%"PRIu64" %"PRIu64"\n", i->start, i->end);
	fprintf(fp, "%d\n", a->thread);
	fprintf(fp, "%d\n", a->phase);
	fprintf(fp, "%d\n", a->access);
	fprintf(fp, "%s\n", a->sig);
	fprintf(fp, "%d\n", myt == a->thread ? 1 : 0);
}

#if 0
void write_race_info(shared race_info_t *info, FILE *fp) {
	fprintf(fp, "%ld\n", (long) info->addr);
	fprintf(fp, "%d\n", info->thread);
	fprintf(fp, "%d\n", info->phase);
	fprintf(fp, "%d\n", info->access);
	// Skipping lockset
	fprintf(fp, "%s\n", info->sig);
	fprintf(fp, "%d\n", info->owned);
}
#endif

void read_race_info(race_info_t *info, FILE* fp) {
	char buf[MAX_LINE_SIZE];
	char *p;
    char *rc;
	int n;
	rc = fgets(buf, MAX_LINE_SIZE, fp);
	info->addr[0] = strtoull(buf, &p, 10);
	info->addr[1] = strtoull(p, NULL, 10);
	//fprintf(stderr, "ADDR = %p\n", info->addr);	// debug

	rc = fgets(buf, MAX_LINE_SIZE, fp);
	info->thread = atoi(buf);
	//fprintf(stderr, "THREAD = %d\n", info->thread);	// debug

	rc = fgets(buf, MAX_LINE_SIZE, fp);
	info->phase = atoi(buf);

	rc = fgets(buf, MAX_LINE_SIZE, fp);
	info->access = atoi(buf);
	//fprintf(stderr, "ACCESS = %s\n", info->access == READ ? "R" : "W");	// debug

	info->locklist = 0;	// Skipping lockset
	//locklist_print(info->locklist);	// debug

	rc = fgets(buf, MAX_LINE_SIZE, fp);
	n = strlen(buf);
	buf[n-1] = 0;	// remove newline
	////fprintf(stderr, "FILE(local) = %s\n", buf);	// debug
	info->sig = strdup(buf);
#if 0
	/*
    //= (shared [] char*) upc_alloc(n);
    //upc_memput(info->file, buf, n);
    ls_strcpy(buf, info->file);
	 */
	//fprintf(stderr, "FILE = %s\n", info->file);	// debug

	rc = fgets(buf, MAX_LINE_SIZE, fp);
	n = strlen(buf);
	buf[n-1] = 0;	// remove newline

	tmp = (char*) malloc(n);
	strncpy(tmp, buf, n);
	info->expr = tmp;

	/*
    //info->expr =  (shared [] char*) upc_alloc(n);
    //upc_memput(info->expr, buf, n);
    ls_strcpy(buf, info->expr);
	 */
	//fprintf(stderr, "EXPR = %s\n", info->expr);	// debug

	rc = fgets(buf, MAX_LINE_SIZE, fp);
	info->line = atoi(buf);
	//fprintf(stderr, "LINE = %d\n", info->line);	// debug
#endif

	rc = fgets(buf, MAX_LINE_SIZE, fp);
	info->owned = atoi(buf);
	//fprintf(stderr, "OWNED = %s\n", (info->owned == TRUE) ? "Y" : "N");	// debug
}

void write_race_pair(int id, int myt, addr_interval* i1, access_info* a1,
		addr_interval*i2, access_info*a2)
{
	sprintf(fname, "%s.%d", RACE_INFO_FILENAME, id);
	FILE* fp = fopen(fname, "w");
	write_race_info(fp, i1, a1, myt);
	write_race_info(fp, i2, a2, myt);
	fclose(fp);
}

#if 0
void write_race_pair(shared race_info_t *i1, 
		shared race_info_t *i2, int id) {

	sprintf(fname, "%s.%d", RACE_INFO_FILENAME, id);
	FILE* fp = fopen(fname, "w");
	write_race_info(i1, fp);
	write_race_info(i2, fp);
	fclose(fp);
}
#endif

void read_race_pair(race_info_t *i1, 
		race_info_t *i2, int id) {

	sprintf(fname, "%s.%d", RACE_INFO_FILENAME, id);
	FILE* fp = fopen(fname, "r");
	if(fp) {
		read_race_info(i1, fp);
		read_race_info(i2, fp);
		fclose(fp);
	}
	else {
		fprintf(stderr, "UPCT: cannot read race pair file %s\n", fname);
		bupc_exit(-1);
	}
}

void print_race_data(lock_pool* lp, race_data* rd)
{
	fprintf(stderr, "RaceData (size=?) { [%"PRIx64",%"PRIx64"), %s, ", /*rd->size,*/ rd->addr.start, rd->addr.end,
			rd->access == READ ? "READ" : "WRITE");
	lock_set_print(lp, &(rd->ls));
	fprintf(stderr, "}\n");
}

void print_access(addr_interval* mem, access_info* ai, int myt) {
    if(ai->thread == MULTIPLE_THREADS) {
        fprintf(stderr, "[%d] %s [0x%"PRIx64",0x%"PRIx64") by thread * at phase %d (%s@%p)\n",
                myt, ai->access == READ ? "Read from" : "Write to",
                mem->start, mem->end, ai->phase, ai->sig, ai->sig);
    }
    else {
        fprintf(stderr, "[%d] %s [0x%"PRIx64",0x%"PRIx64") by thread %d at phase %d (%s@%p)\n",
                myt, ai->access == READ ? "Read from" : "Write to",
                mem->start, mem->end, ai->thread, ai->phase, ai->sig, ai->sig);
    }
}

void print_racedata(race_data* rd) {
	fprintf(stderr, "%s [0x%"PRIx64",0x%"PRIx64") (%s@%p)\n",
			rd->access == READ ? "Read from" : "Write to",
					rd->addr.start, rd->addr.end, rd->sig, rd->sig);
}

void open_stat(upct_stat* stat, int myt)
{
	sprintf(fname, "%s.%d", STAT_FILENAME, myt);
	if (!(stat->stat_file = fopen(fname, "w"))) {
		fprintf(stderr, "Warning: UPCT: cannot open %s for writing statistics\n", fname);
	}
	else {
		fprintf(stat->stat_file, "Phase:Shared_reads,Shared_writes,Stronger_reads,"
				"Stronger_writes,ISTree_size,Bytes_sent,Bytes_received\n");
	}
}

void write_stat(upct_stat* stat, int phase)
{
	if (stat->stat_file) {
		fprintf(stat->stat_file, "%d:%d,%d,%d,%d,%d,%d,%d\n", phase, stat->n_shread,
				stat->n_shwrite, stat->n_strongerread, stat->n_strongerwrite,
				stat->sz_istree, stat->n_bytesent, stat->n_byterecv);
	}
}

void close_stat(upct_stat* stat)
{
	if (stat->stat_file)
		fclose(stat->stat_file);
}


