/*
 * upct_types.h - macros, typedefs, structs for upc thrille
 *
 */

#ifndef UPCT_TYPES_H
#define UPCT_TYPES_H

#include <stdio.h>
#include <upc.h>
#include <assert.h>
#include <inttypes.h>
#include "lockset/lockset.h"
#include "access.h"
#include "intervals/interval.h"

//#define TRACE_FILENAME	"thrille.trace"
//#define TRACE_FILENAME_SIZE	256	

#define THRILLE_VERSION "0.9.17"

// Bit flags to turn on debugging
#define THRILLE_DEBUG_ACCESS	1
#define THRILLE_DEBUG_LOCK		2
#define THRILLE_DEBUG_BARRIER	4
#define THRILLE_DEBUG_VERBOSE	8

#define THRILLE_BEGIN_FUNC(x)	if (!thrille_info.initialized || thrille_info.inside_inst) \
                                    return x;\
                                    thrille_info.inside_inst++

#define THRILLE_END_FUNC	thrille_info.inside_inst--
#define THRILLE_LOCK		upcr_lock(thrille_lock)
#define THRILLE_UNLOCK		upcr_unlock(thrille_lock)

#define MAX_SEM_TRIES	100
#define SEM_WAIT_USEC   1000

#define REPORTED 1
#define NOT_REPORTED 0

#define RACE 1
#define NO_RACE -1
#define BAD_RACE -2

#define EMPTY 0
#define NONEMPTY 1

#define VALID 1
#define INVALID 0

#define FOUND 1
#define NOT_FOUND 0

typedef int BOOL;
#define FALSE 0
#define TRUE 1

#ifndef BACK_OFF
#   define BACK_OFF	(0.95)
#endif

#ifndef EPS
#    define EPS			(1.0/RAND_MAX)
#endif

typedef uint64_t lock_addr;
typedef uint64_t mem_addr;


struct _race_info_t;
struct _race_target_t;
struct _thrille_info_t;
struct _list_elem;


typedef struct _race_info_t race_info_t;
typedef struct _race_target_t race_target_t;
typedef struct _thrille_info_t thrille_info_t;
typedef struct _list_elem list_elem;

//typedef bupc_sem_t* ptr_sem;


struct _race_info_t {
    mem_addr addr[2];
    int thread;
    enum _AccessType access;
    shared list_elem * locklist; 
    const char* sig;
    int race_reported;
    unsigned int phase;
    int owned;
};

struct _lockset_t;

struct _thrille_info_t {
    int initialized;
    int inside_inst;
    shared struct _set_t * ls;
    unsigned int phase;
};

typedef struct _mem_range_t {
    uint64_t start;
    uint64_t end;
} mem_range;

#if 0
/* Change this struct to a more generic bit-field for flexibity
 * in number of locks allowed / locks held in program
 */ 
typedef struct _lock_set_t {
    uint32_t idx;
    uint32_t bits;
} lock_set;
#endif

/* Minimum data required to send off to other threads for computation
 * of potential races -- compact as much as possible
 */
typedef struct _race_data_t {
//    int    size;
    mem_range   addr;
    enum _AccessType access;
    const char*	sig;
    lock_set    ls;		// must be last element
} race_data;


typedef struct _race_data_list_t {
	int rd_size;
    race_data* rd;
    struct _race_data_list_t* next;
} race_data_list;

#define MEM_ACC_BLOCK_SIZE	64*1024	// observed max was 47K
#define MAXTHREADS	256

typedef struct _mem_access_pp_t {
    int activity;
    char in[THREADS][MEM_ACC_BLOCK_SIZE];
    char out[THREADS][MEM_ACC_BLOCK_SIZE];
} mem_access_pp;

#define THRILLE_PHASE_HISTORY   4   // Must be a power of 2
#define thrille_phase_idn(x)    ((x) & (THRILLE_PHASE_HISTORY-1))   
#define thrille_phase_id        thrille_phase_idn(thrille_info.phase)

typedef struct _mem_access_db_t {
    mem_access_pp pp[THRILLE_PHASE_HISTORY];
} mem_access_db;


typedef struct _upct_stat_t {
	FILE* stat_file;
	int n_shread;
	int n_shwrite;
	//int n_barrier;
	int n_strongerread;
	int n_strongerwrite;
	int sz_istree;
	int n_bytesent;
	int n_byterecv;
} upct_stat;

#define clear_stat(x)	do {\
	x.n_shread = x.n_shwrite = x.n_strongerread = x.n_strongerwrite = x.sz_istree =\
	x.n_bytesent = x.n_byterecv = 0; } while(0)

void print_race_info(shared race_info_t *);
void print_race_info_ns(race_info_t *);
void write_race_pair(int, int, addr_interval*, access_info*, addr_interval*, access_info*);
void read_race_pair(race_info_t *, race_info_t *, int);
void print_race_data(lock_pool*, race_data*);
void print_access(addr_interval* mem, access_info* ai, int myt);
void print_racedata(race_data* rd);

void open_stat(upct_stat* stat, int myt);
void write_stat(upct_stat* stat, int phase);
void close_stat(upct_stat* stat);

#endif // UPCT_TYPES_H
