/* hello-bulk.upc
 * Authors - Chang-Seo Park (parkcs@cs.berkeley.edu)
 *         - Nick Jalbert   (jalbert@cs.berkeley.edu)
 *
 * Tests cast-to-local races
 */

#include <upc.h>
#include <stdio.h>
#include <assert.h>

#pragma relaxed


shared int * sum_ptr;


int main(int argc, char* argv[])
{
    sum_ptr = (shared int *) upc_all_alloc(1, sizeof(int));
    upc_lock_t* the_lock = upc_all_lock_alloc();

    upc_barrier;

    if (upc_threadof(sum_ptr) == MYTHREAD) {
        printf("[%d] casting to local\n", MYTHREAD);
        int * p;
        p = (int *) sum_ptr;
        //thrille_before_local_write(MYTHREAD, (long) p, __FILE__, __LINE__, "*p = THREADS+1");
        
        // This is a racy access
        *p = THREADS+1;
    } else {
        upc_lock(the_lock);
        *sum_ptr = MYTHREAD;
        upc_unlock(the_lock);
    }

    if (upc_threadof(sum_ptr) == MYTHREAD) {
        printf("[%d] casting to local a second time\n", MYTHREAD);
        int * p;
        p = (int *) sum_ptr;
        //thrille_before_local_write(MYTHREAD, (long) p, __FILE__, __LINE__, "*p = THREADS+10");
        
        // This is another racy access
        *p = THREADS+10;
    }

    upc_barrier;

    if (MYTHREAD == 0) {
        printf("Last thread to assign is %d\n", *sum_ptr);
        // Potential bug if local assignment is the last assignment
        assert(*sum_ptr >= 0 && *sum_ptr <= THREADS);
        upc_lock_free(the_lock);
    }

    return 0;
}


