#define N   8 /* Dimension of vectors/matrices */

shared [N] int a[N][N];
shared int b[N];
shared int c[N];

/* C = A \dot B */
void vecTvec(shared int A[], shared int B[], shared int* C) {
    *C = 0;
    for(int i = 0; i < N; i++)
        *C += A[i]*B[i];
}

void matvec(shared [N] int A[N][N], shared int B[N], shared int C[N]) {
    int b[N];
    int c[N];
    for(int i = 0; i < N; i++) {
        b[i] = B[i];
    }

    upc_notify;     // nofity "owner" that it is okay to update

    upc_forall(int i = 0; i < N; i++; &C[i]) {
        //printf("Compute iteration %d by thread %d\n", i, MYTHREAD);

        //Using local var
        c[i] = 0;
        for(int j = 0; j < N; j++)
            c[i] += A[i][j] * b[j];

    }

    upc_wait;   // wait before actually writing

    upc_forall(int i = 0; i < N; i++; &C[i]) {
        C[i] = c[i];
    }
}

void printvec(shared int A[]) {
    printf("( ");
    for(int i = 0; i < N; i++) {
        printf("%2d ", A[i]);
    }
    printf(")");
}

int main(int argc, char* argv[])
{
    srand(MYTHREAD * 8);
    upc_forall(int i = 0; i < N; i++; &a[i]) {
        for(int j = 0; j < N; j++) {
            a[i][j] = rand() % 10;
        }
    }
    upc_forall(int i = 0; i < N; i++; &b[i]) {
        //printf("Init iteration %d by thread %d\n", i, MYTHREAD);
        b[i] = rand() % 10;
        c[i] = 0;
    }

    upc_barrier;

    if (MYTHREAD == 0) {
        for(int x = 0; x < N; x++) {
            for(int y = 0; y < N; y++) {
                printf("%2d ", a[x][y]);
            }
            printf("\n");
        }

        printf(" . ");
        printvec(b);
        printf("T");
    }

    upc_barrier;

    matvec(a, b, b);
    //upc_barrier;
    matvec(a, b, b);

    upc_barrier;

    if(MYTHREAD == 0) {
        printf(" = ");
        printvec(b);
        printf("T\n");
    }

    upc_barrier;
}
