#define N   8 /* Dimension of vectors/matrices */

shared [N] double a[N][N];
shared double b[N];
shared double c[N];

/* C = A \dot B */
void vecTvec(shared double A[], shared double B[], shared double* C) {
    *C = 0;
    for(int i = 0; i < N; i++)
        *C += A[i]*B[i];
}

void matvec(shared [N] double A[N][N], shared double B[N], shared double C[N]) {
    double b[N];
    for(int i = 0; i < N; i++) {
        b[i] = B[i];
    }
    upc_forall(int i = 0; i < N; i++; &C[i]) {
       //printf("Compute iteration %d by thread %d\n", i, MYTHREAD);

       //Using local var
       double c = 0.0;
       for(int j = 0; j < N; j++)
           c += A[i][j] * b[j];
        C[i] = c;

       /* 
       // Using "naive" add
       C[i] = 0.0;
       for(int j = 0; j < N; j++)
           C[i] += A[i][j] * B[j];
        */
    }
}

void printvec(shared double A[]) {
    printf("( ");
    for(int i = 0; i < N; i++) {
        printf("%f ", A[i]);
    }
    printf(")");
}

int main(int argc, char* argv[])
{
   srand(MYTHREAD * 8);
   upc_forall(int i = 0; i < N; i++; &a[i]) {
       for(int j = 0; j < N; j++) {
            a[i][j] = rand() % 10;
       }
   }
   upc_forall(int i = 0; i < N; i++; &b[i]) {
       //printf("Init iteration %d by thread %d\n", i, MYTHREAD);
       b[i] = rand() % 10;
       c[i] = 0;
   }

   upc_barrier;

   if (MYTHREAD == 0) {
       for(int x = 0; x < N; x++) {
            for(int y = 0; y < N; y++) {
                printf("%f ", a[x][y]);
            }
           printf("\n");
       }

       printf(" . ");
       printvec(b);
       printf("T");
   }
   
   upc_barrier;

   matvec(a, b, b);

   upc_barrier;

   if(MYTHREAD == 0) {
       printf(" = ");
       printvec(b);
       printf("T\n");
   }

}
